import os
import struct
import time
import zlib
import hashlib

def crc32_of(data: bytes) -> int:
    """
    Compute CRC-32 of 'data'.
    """
    return zlib.crc32(data) & 0xffffffff


# =============================================================================
# 1) Ping Request (295 bits)
#    - 256-bit nonce
#    - 7-bit version
#    - 32-bit CRC
#    = 295 bits total
#    In practice, we store 37 bytes (296 bits); 1 bit is unused.
# =============================================================================

def build_ping_request(version: int) -> bytes:
    """
    Build a Ping request with:
      - 256-bit nonce (32 bytes)
      - 7-bit version
      - 32-bit CRC
    We do bit-packing. The final result is 37 bytes (296 bits), with 1 unused bit.
    """
    if not (0 <= version < 128):
        raise ValueError("Version must fit in 7 bits (0..127)")

    # 1) Generate 256-bit nonce
    nonce = os.urandom(32)  # 32 bytes = 256 bits

    # We'll build partial_data = [nonce (256 bits), version (7 bits)] as an integer
    # Then compute CRC-32 over those bytes, then append 32 bits of CRC.
    partial_int = int.from_bytes(nonce, 'big') << 7  # shift left 7 bits
    partial_int |= version  # put version in the low 7 bits

    # Convert partial to bytes
    # partial is 256+7=263 bits => needs 33 bytes to store
    partial_bytes = partial_int.to_bytes(33, 'big')

    # Compute CRC over partial_bytes
    cval = crc32_of(partial_bytes)

    # Now combine partial_data (263 bits) with 32 bits of CRC => 295 bits
    # We'll store that in a single integer
    final_int = (int.from_bytes(partial_bytes, 'big') << 32) | cval
    # final_int is 263+32=295 bits, needs 37 bytes to store (the last bit is unused).
    final_bytes = final_int.to_bytes(37, 'big')
    return final_bytes


def parse_ping_request(data: bytes):
    """
    Parse a Ping request (37 bytes = 295 bits).
    Returns (nonce, version) or None if invalid.
    """
    if len(data) != 37:
        return None

    # Convert to int
    val_295 = int.from_bytes(data, 'big')  # 295 bits in a 37-byte integer
    # Extract CRC (lowest 32 bits)
    crc_in = val_295 & 0xffffffff
    # Then shift right 32 bits to get partial_data
    partial_val = val_295 >> 32  # 263 bits

    # Convert partial_val back to bytes
    partial_bytes = partial_val.to_bytes(33, 'big')

    # Recompute CRC
    crc_calc = crc32_of(partial_bytes)
    if crc_calc != crc_in:
        return None

    # Now parse out nonce (256 bits) and version (7 bits)
    # partial_val is 263 bits
    version = partial_val & 0x7f  # low 7 bits
    nonce_val = partial_val >> 7  # high 256 bits
    nonce_bytes = nonce_val.to_bytes(32, 'big')

    return (nonce_bytes, version)


# =============================================================================
# 2) Ping Response (72 bits)
#    - 32-bit timestamp
#    - 7-bit version + 1-bit answer => 8 bits
#    - 32-bit CRC
#    = 72 bits total => 9 bytes
# =============================================================================

def build_ping_response(version: int, answer: int) -> bytes:
    """
    Build a Ping response:
      - 32-bit timestamp (lowest 32 bits of current time in ms)
      - 7-bit version + 1-bit answer
      - 32-bit CRC
    => 72 bits = 9 bytes
    """
    if not (0 <= version < 128):
        raise ValueError("Version must fit in 7 bits.")
    if answer not in (0, 1):
        raise ValueError("Answer must be 0 or 1.")

    # 32-bit timestamp = current time in ms, truncated to 32 bits
    t_ms = int(time.time() * 1000) & 0xffffffff

    # partial = [timestamp (32 bits), version (7 bits), answer (1 bit)] => 40 bits
    partial_val = (t_ms << 8) | ((version << 1) & 0xfe) | (answer & 0x01)
    # partial_val is 40 bits => 5 bytes
    partial_bytes = partial_val.to_bytes(5, 'big')

    # CRC over these 5 bytes
    cval = crc32_of(partial_bytes)

    # Combine partial (40 bits) with 32 bits of CRC => 72 bits total
    final_val = (int.from_bytes(partial_bytes, 'big') << 32) | cval
    final_bytes = final_val.to_bytes(9, 'big')
    return final_bytes


def parse_ping_response(data: bytes):
    """
    Parse a Ping response (72 bits = 9 bytes).
    Return (timestamp_ms, version, answer) or None if invalid.
    """
    if len(data) != 9:
        return None

    val_72 = int.from_bytes(data, 'big')  # 72 bits
    crc_in = val_72 & 0xffffffff
    partial_val = val_72 >> 32  # 40 bits

    partial_bytes = partial_val.to_bytes(5, 'big')
    crc_calc = crc32_of(partial_bytes)
    if crc_calc != crc_in:
        return None

    # Now parse partial_val
    # partial_val = [timestamp(32 bits), version(7 bits), answer(1 bit)]
    t_ms = (partial_val >> 8) & 0xffffffff
    va = partial_val & 0xff  # 8 bits = [7 bits version, 1 bit answer]
    version = (va >> 1) & 0x7f
    answer = va & 0x01

    return (t_ms, version, answer)


# =============================================================================
# 3) Handshake
#    - 32-bit timestamp
#    - 64-byte ephemeral pubkey (raw x||y = 512 bits)
#    - 64-byte ephemeral signature (raw r||s = 512 bits)
#    - 32-byte PFS hash (256 bits)
#    - 32-bit CRC
#    => total 4 + 64 + 64 + 32 + 4 = 168 bytes = 1344 bits
# =============================================================================

def build_handshake_message(timestamp: int,
                            ephemeral_pubkey: bytes,
                            ephemeral_signature: bytes,
                            pfs_hash: bytes) -> bytes:
    """
    Build handshake:
      - 4 bytes: timestamp
      - 64 bytes: ephemeral_pubkey (x||y, raw)
      - 64 bytes: ephemeral_signature (r||s, raw)
      - 32 bytes: pfs_hash
      - 4 bytes: CRC-32
    => 168 bytes total
    """
    if len(ephemeral_pubkey) != 64:
        raise ValueError("ephemeral_pubkey must be 64 bytes (raw x||y).")
    if len(ephemeral_signature) != 64:
        raise ValueError("ephemeral_signature must be 64 bytes (raw r||s).")
    if len(pfs_hash) != 32:
        raise ValueError("pfs_hash must be 32 bytes.")

    partial = struct.pack("!I", timestamp) \
              + ephemeral_pubkey \
              + ephemeral_signature \
              + pfs_hash
    cval = crc32_of(partial)
    return partial + struct.pack("!I", cval)


def parse_handshake_message(data: bytes):
    """
    Parse handshake message (168 bytes).
    Return (timestamp, ephemeral_pub, ephemeral_sig, pfs_hash) or None if invalid.
    """
    if len(data) != 168:
        return None
    partial = data[:-4]   # first 164 bytes
    crc_in = struct.unpack("!I", data[-4:])[0]
    crc_calc = crc32_of(partial)
    if crc_calc != crc_in:
        return None

    # Now parse fields
    timestamp = struct.unpack("!I", partial[:4])[0]
    ephemeral_pub = partial[4:4+64]
    ephemeral_sig = partial[68:68+64]
    pfs_hash = partial[132:132+32]
    return (timestamp, ephemeral_pub, ephemeral_sig, pfs_hash)


# =============================================================================
# 4) PFS Hash Helper
#    If no previous session, return 32 zero bytes
#    Otherwise, compute sha256(session_number || last_shared_secret).
# =============================================================================

def compute_pfs_hash(session_number: int, shared_secret_hex: str) -> bytes:
    """
    Return 32 bytes (256 bits) for the PFS field.
    If session_number < 0 => means no previous session => 32 zero bytes.
    Otherwise => sha256( session_number (4 bytes) || shared_secret ).
    """
    if session_number < 0:
        return b"\x00" * 32

    # Convert shared_secret_hex to raw bytes
    secret_bytes = bytes.fromhex(shared_secret_hex)
    # Pack session_number as 4 bytes
    sn_bytes = struct.pack("!I", session_number)
    return hashlib.sha256(sn_bytes + secret_bytes).digest()
